""" GenerateBarcodeWindow Class """

import cv2
import tkinter
import tkinter.filedialog
from tkinter.messagebox import showerror, showwarning, showinfo
import copy
import os
import threading

from kalmus.tkinter_windows.meta_info_windows.SpecifyMetaDataWindow import SpecifyMetaDataWindow
from kalmus.tkinter_windows.gui_utils import resource_path


class GenerateBarcodeWindow():
    """
    GenerateBarcodeWindow Class
    GUI window for user to generate the barcode from a video file
    """

    def __init__(self, barcode_generator, barcode_stack):
        """
        Initialize

        :param barcode_generator: The barcode generator
        :param barcode_stack: The dictionary that stores all the barcode on the memory
        """
        self.barcode_generator = barcode_generator
        self.barcode_stack = barcode_stack

        # Initialize the window
        self.window = tkinter.Tk()
        self.window.wm_title("Barcode Generator")
        self.window.iconbitmap(resource_path("kalmus_icon.ico"))

        # A temporary meta data dictionary that will hold the meta information given by the user
        self.meta_data_dict = {}

        # video object capture by cv2
        self.video = None

        # Label prompt for the generated barcode's barcode type
        barcode_type_label = tkinter.Label(self.window, text="Barcode Type: ")
        barcode_type_label.grid(row=0, column=0)

        # Label prompt for the frame sampling type
        frame_type_label = tkinter.Label(self.window, text="Frame Type: ")
        frame_type_label.grid(row=1, column=0)

        # Label prompt for the color/brightness metric
        color_metric_label = tkinter.Label(self.window, text="Color metric: ")
        color_metric_label.grid(row=2, column=0)

        # Variable that stores the user's choice of generated barcode type
        self.barcode_type_var = tkinter.StringVar(self.window)
        self.barcode_type_var.set("Color")

        # Dropdown menu for the barcode type selection
        dropdown_bar_type = tkinter.OptionMenu(self.window, self.barcode_type_var, "Color", "Brightness")
        dropdown_bar_type.grid(row=0, column=1)

        # Variable that stores the user's choice of frame sampling type
        self.frame_type_var = tkinter.StringVar(self.window)
        self.frame_type_var.set("Whole_frame")

        # Dropdown menu for the frame sampling type selection
        dropdown_frame_type = tkinter.OptionMenu(self.window, self.frame_type_var, "Whole_frame", "Low_contrast_region",
                                                 "High_contrast_region", "Foreground", "Background")
        dropdown_frame_type.grid(row=1, column=1)

        # Variable that stores the user's choice of color metric
        self.color_metric_var = tkinter.StringVar(self.window)
        self.color_metric_var.set("Average")

        # Dropdown menu for the color metric selection
        dropdown_color_metric = tkinter.OptionMenu(self.window, self.color_metric_var, "Average", "Median", "Mode",
                                                   "Top-dominant", "Weighted-dominant", "Bright", "Brightest")
        dropdown_color_metric.grid(row=2, column=1)

        # Label prompt for the skip over specification
        self.skip_over_label = tkinter.Label(self.window, text="Start at (frames): ")
        self.skip_over_label.grid(row=0, column=2)

        # Label prompt for the sample rate specification
        self.sampled_rate_label = tkinter.Label(self.window, text="Sample every (frames): ")
        self.sampled_rate_label.grid(row=1, column=2)

        # Label prompt for the total frame included specification
        self.total_frames_label = tkinter.Label(self.window, text="Total frames: ")
        self.total_frames_label.grid(row=2, column=2)

        # Acquisition unit specification (in frames or in time)
        self.acquisition_label = tkinter.Label(self.window, text="Acquistion unit")
        self.acquisition_label.grid(row=0, column=5)

        # Variable that stores the user's choice of acquisition unit
        self.acquisition_option = tkinter.StringVar(self.window)
        self.acquisition_option.set("Frame")

        # Radio button for the Acquisition unit (Frame/Time) selection
        radio_frame = tkinter.Radiobutton(self.window, text="Frame", variable=self.acquisition_option,
                                          value="Frame", anchor='w',
                                          command=self.frame_unit)
        radio_frame.grid(row=1, column=5, sticky=tkinter.W)
        radio_frame.select()

        radio_time = tkinter.Radiobutton(self.window, text="Time", variable=self.acquisition_option,
                                         value="Time", anchor='w',
                                         command=self.time_unit)
        radio_time.grid(row=2, column=5, sticky=tkinter.W)

        # Text entry for the skip over specification
        self.skip_over_entry = tkinter.Entry(self.window, textvariable="0", width=12)
        self.skip_over_entry.grid(row=0, column=3, columnspan=2)

        # Text entry for the sampling rate specification
        self.sampled_rate_entry = tkinter.Entry(self.window, textvariable="1", width=12)
        self.sampled_rate_entry.grid(row=1, column=3, columnspan=2)

        # Text entry for the total frames specification
        self.total_frames_entry = tkinter.Entry(self.window, textvariable="1000", width=12)
        self.total_frames_entry.grid(row=2, column=3, columnspan=2)

        # Label prompt for the video file name input
        video_filename_label = tkinter.Label(self.window, text="Media file path: ")
        video_filename_label.grid(row=3, column=0, columnspan=1)

        # Text entry for the video file name specification
        self.video_filename_entry = tkinter.Entry(self.window, textvariable="", width=55)
        self.video_filename_entry.grid(row=3, column=1, columnspan=3)

        # Button to browse the folder
        self.browse_folder_button = tkinter.Button(self.window, text='Browse', command=self.browse_folder)
        self.browse_folder_button.grid(row=3, column=4)

        # Variable that stores the letter box option (automatic/user defined)
        self.letterbox_option = tkinter.StringVar(self.window)
        self.letterbox_option.set("Auto")  # initialize

        # Label prompt for the letter box remove option
        letterbox_label = tkinter.Label(self.window, text="Remove Letterbox: ")
        letterbox_label.grid(row=4, column=0, columnspan=1)

        # Radio button for the letter box remove option
        radio_auto = tkinter.Radiobutton(self.window, text="Auto", variable=self.letterbox_option,
                                         value="Auto", anchor='w',
                                         command=self.disable_setup)
        radio_auto.grid(row=5, column=0, sticky=tkinter.W)
        radio_auto.select()

        radio_manual = tkinter.Radiobutton(self.window, text="Manual", variable=self.letterbox_option,
                                           value="Manual", anchor='w',
                                           command=self.enable_setup)
        radio_manual.grid(row=6, column=0, sticky=tkinter.W)

        # User defined letter box label prompt
        high_ver_label = tkinter.Label(self.window, text="Upper vertical: ")
        high_ver_label.grid(row=5, column=1, columnspan=1)

        low_ver_label = tkinter.Label(self.window, text="Lower vertical: ")
        low_ver_label.grid(row=6, column=1, columnspan=1)

        # User defined letter box text entry
        self.high_ver_entry = tkinter.Entry(self.window, textvariable="-1", width=4, state="disabled")
        self.high_ver_entry.grid(row=5, column=2, columnspan=1, sticky=tkinter.W)

        self.low_ver_entry = tkinter.Entry(self.window, textvariable="-2", width=4, state="disabled")
        self.low_ver_entry.grid(row=6, column=2, columnspan=1, sticky=tkinter.W)

        left_hor_label = tkinter.Label(self.window, text="Left horizontal: ")
        left_hor_label.grid(row=5, column=3, columnspan=1)

        right_hor_label = tkinter.Label(self.window, text="right horizontal: ")
        right_hor_label.grid(row=6, column=3, columnspan=1)

        self.left_hor_entry = tkinter.Entry(self.window, textvariable="-3", width=4, state="disabled")
        self.left_hor_entry.grid(row=5, column=4, columnspan=1)

        self.right_hor_entry = tkinter.Entry(self.window, textvariable="-4", width=4, state="disabled")
        self.right_hor_entry.grid(row=6, column=4, columnspan=1)

        # Variable that stores 0 for not saving frames during the generation 1 for saving frames during the generation
        self.var_saved_frame = tkinter.IntVar(self.window)

        # Checkbox for the user to choose whether save the frames or not during barcode generation
        self.checkbox_saved_frame = tkinter.Checkbutton(self.window, text="Save Frames  ",
                                                        variable=self.var_saved_frame,
                                                        onvalue=1, offvalue=0, command=self.update_save_frame_entry)
        self.checkbox_saved_frame.grid(row=7, column=0)

        # Label prompt for saving frame
        save_frame_label = tkinter.Label(master=self.window, text="Save every (secs):")
        save_frame_label.grid(row=7, column=1, sticky=tkinter.E)

        # Text entry for the saved frames rate specification
        self.save_frame_entry = tkinter.Entry(self.window, textvariable="-8", width=4, state="normal")
        self.save_frame_entry.grid(row=7, column=2, sticky=tkinter.W)
        self.save_frame_entry.delete(0, tkinter.END)
        self.save_frame_entry.insert(0, "4")
        self.save_frame_entry.config(state="disabled")

        # Variable that stores 0 for not saving frames during the generation 1 for rescaling frame
        # during the generation
        self.var_rescale_frame = tkinter.IntVar(self.window)

        # Checkbox for the user to choose whether to rescale the frames or not during the barcode generation
        self.checkbox_rescale_frame = tkinter.Checkbutton(self.window, text="Rescale Frames", width=12,
                                                          variable=self.var_rescale_frame,
                                                          onvalue=1, offvalue=0, command=self.update_rescale_entry)
        self.checkbox_rescale_frame.grid(row=7, column=3, sticky=tkinter.E)

        # Text entry for the rescale factor specification
        self.rescale_factor_entry = tkinter.Entry(self.window, textvariable="-7", width=4, state="normal")
        self.rescale_factor_entry.grid(row=7, column=4)
        self.rescale_factor_entry.delete(0, tkinter.END)
        self.rescale_factor_entry.insert(0, "0.5")
        self.rescale_factor_entry.config(state="disabled")

        # Variable that stores 0 for single thread generation 1 for multi-thread generation
        self.var_multi_thread = tkinter.IntVar(self.window)

        # Checkbox for the user to choose whether use the multi-thread or not for barcode generation
        self.checkbox_multi_thread = tkinter.Checkbutton(self.window, text="Multi-Thread:",
                                                         variable=self.var_multi_thread,
                                                         onvalue=1, offvalue=0, command=self.update_thread_entry)

        self.checkbox_multi_thread.grid(row=8, column=0)

        # Text entry for the thread specification
        self.thread_entry = tkinter.Entry(self.window, textvariable="-6", width=4, state="normal")
        self.thread_entry.grid(row=8, column=1, sticky=tkinter.W)
        self.thread_entry.delete(0, tkinter.END)
        self.thread_entry.insert(0, "4")
        self.thread_entry.config(state="disabled")

        # Button to generate the barcode
        self.generate_button = tkinter.Button(master=self.window, text="Generate Barcode",
                                              command=self.generate_barcode_thread)
        self.generate_button.grid(row=8, column=2, sticky=tkinter.W, rowspan=1, pady=5)

        # Button to specify the meta data of the generated barcode
        self.specify_data_button = tkinter.Button(master=self.window, text="Specify Meta Data",
                                                  command=self.specify_data)
        self.specify_data_button.grid(row=8, column=3, sticky=tkinter.W, rowspan=1, pady=5)

        self.window.protocol("WM_DELETE_WINDOW", self.quit)

        # Start the window
        self.window.mainloop()

    def quit(self):
        """
        Quit the main window
        """
        self.window.quit()
        self.window.destroy()

    def specify_data(self):
        """
        Instantiate the SpecifyMetaDataWindow
        """
        SpecifyMetaDataWindow(self.meta_data_dict)

    def update_thread_entry(self):
        """
        Enable or disable and change the label prompt when user check/uncheck the multi-thread checkbox
        """
        if self.var_multi_thread.get() == 0:
            self.checkbox_multi_thread["text"] = "Multi-Thread:"
            self.thread_entry.config(state="disabled")
        elif self.var_multi_thread.get() == 1:
            self.checkbox_multi_thread["text"] = "# of Threads:"
            self.thread_entry.config(state="normal")

    def update_rescale_entry(self):
        """
        Enable or disable and change the label prompt when user check/uncheck the rescale frames checkbox
        """
        if self.var_rescale_frame.get() == 0:
            self.checkbox_rescale_frame["text"] = "Rescale Frames"
            self.rescale_factor_entry.config(state="disabled")
        elif self.var_rescale_frame.get() == 1:
            self.checkbox_rescale_frame["text"] = "By a factor of: "
            self.rescale_factor_entry.config(state="normal")

    def update_save_frame_entry(self):
        """
        Enable or disable the text entry when user check/uncheck the save frames checkbox
        """
        if self.var_saved_frame.get() == 0:
            self.save_frame_entry.config(state="disabled")
        elif self.var_saved_frame.get() == 1:
            self.save_frame_entry.config(state="normal")

    def time_unit(self):
        """
        Change the acquisition unit to Time when user switch to the time radio button
        """
        self.skip_over_label['text'] = "Start at (mins:secs): "
        self.sampled_rate_label['text'] = "Sample every (secs): "
        self.total_frames_label['text'] = "End at (mins:secs): "

    def frame_unit(self):
        """
        Change the acquisition unit to Frame when user switch to the frame radio button
        """
        self.skip_over_label['text'] = "Start at (frames): "
        self.sampled_rate_label['text'] = "Sample every (frames): "
        self.total_frames_label['text'] = "Total frames: "

    def browse_folder(self):
        """
        Browse the folder
        """
        # Get the file name from the user specification
        filename = tkinter.filedialog.askopenfilename(initialdir=".", title="Select Media file",
                                                      filetypes=(("mp4 files", "*.mp4"), ("avi files", "*.avi"),
                                                                 ("m4v files", "*.m4v"), ("All files", "*.*")))

        # Update the file name to the text entry
        self.video_filename_entry.delete(0, tkinter.END)
        self.video_filename_entry.insert(0, filename)

    def generate_barcode_thread(self):
        """
        Generate the barcode in a another thread
        to avoid the frozen tkinter window issue

        :return:
        """
        threading.Thread(target=self.generate_barcode).start()

    def acquire_generation_param(self):
        """
        Acquire the barcode generation parameters
        """
        # Get barcode type, frame sampling type, and color/brightness metric
        barcode_type = self.barcode_type_var.get()
        frame_type = self.frame_type_var.get()
        color_metric = self.color_metric_var.get()

        # Get the acquisition unit
        unit_type = self.acquisition_option.get()

        # Get the file name/path of the input video
        video_filename = str(self.video_filename_entry.get())

        if not os.path.exists(video_filename):
            showerror("Video File Not Exists", "Video file not found!\nPlease check the file path.")
            raise FileNotFoundError()

        # Get the correct acquisition parameters based on the acqusition unit
        if unit_type == "Frame":
            skip_over_str = self.skip_over_entry.get()
            if len(skip_over_str) == 0 or skip_over_str.lower() == "start":
                skip_over = 0
            else:
                skip_over = int(self.skip_over_entry.get())

            total_frames_str = self.total_frames_entry.get()
            if len(total_frames_str) == 0 or total_frames_str.lower() == "end":
                total_frames = int(1e8)
            else:
                total_frames = int(self.total_frames_entry.get())

            sampled_frame_rate_str = self.sampled_rate_entry.get()
            if len(sampled_frame_rate_str) == 0:
                showwarning("Sample Frame not Specified", "Sample frame rate is not given.\n"
                                                          "Default sample rate 1 frame every input frame is used.")
                sampled_frame_rate = 1
            else:
                sampled_frame_rate = int(sampled_frame_rate_str)
        elif unit_type == "Time":
            self.video = cv2.VideoCapture(video_filename)
            fps = self.video.get(cv2.CAP_PROP_FPS)

            skip_over_str = str(self.skip_over_entry.get())
            if len(skip_over_str) == 0 or skip_over_str.lower() == "start":
                skip_over = 0
            else:
                split_pos = skip_over_str.find(":")
                skip_over = int((int(skip_over_str[:split_pos]) * 60 + int(skip_over_str[split_pos + 1:])) * fps)

            sampled_frame_rate_str = str(self.sampled_rate_entry.get())
            if len(sampled_frame_rate_str) == 0:
                showwarning("Sample Frame not Specified", "Sample frame rate is not given.\n"
                                                          "Default sample rate 1 frame every input frame is used.")
                sampled_frame_rate = 1
            else:
                sampled_frame_rate = int(round(float(sampled_frame_rate_str) * fps))

            total_frames_str = str(self.total_frames_entry.get())
            if len(total_frames_str) == 0 or total_frames_str.lower() == "end":
                total_frames = int(1e8)
            else:
                split_pos = total_frames_str.find(":")
                total_frames = (int(total_frames_str[:split_pos]) * 60 + int(total_frames_str[split_pos + 1:])) * fps
                total_frames -= skip_over
                total_frames = int(total_frames)
                total_frames //= sampled_frame_rate

        # Make sure the sampled frame rate >= 1 and the skip over >= 0 and total frame >= 0
        if sampled_frame_rate < 1:
            showwarning("Frame Sample Rate too Small", "The frame sample rate is too small.\n"
                                                       "It has been adjusted to the Minimum valid sample rate,\n"
                                                       "Sample 1 frame every frame (==use all frames)")
            sampled_frame_rate = 1
        if skip_over < 0:
            showwarning("Invalid Start time", "Invalid start time for the barcode generation.\n"
                                              "Start frames/time has been set to 0/00:00 (== start of film)")
            skip_over = 0
        if total_frames < 0:
            showwarning("Invalid Total Frames", "Invalid total frames\n"
                                                "or Barcode starts after it ends.\n"
                                                "Total frames has been adjusted to 0.")
            total_frames = 0

        return barcode_type, frame_type, color_metric, sampled_frame_rate, skip_over, total_frames, video_filename

    def generate_barcode(self):
        """
        Generate the barcode using the given parameters
        """
        self.disable_generate_button()
        try:
            barcode_type, frame_type, color_metric, sampled_frame_rate, skip_over, total_frames, video_filename = \
                self.acquire_generation_param()
        except FileNotFoundError:
            self.enable_generate_button()
            return
        except:
            showerror("Acquisition Parameters", "An unknown Error occurred when reading\n"
                                                "the acquisition parameters.\n\n"
                                                "Please make sure all parameters are positive\n"
                                                "Frames units must all be integers\n"
                                                "In Time unit, mins:secs in Start and End at must be integers\n"
                                                "Sample every (secs) can be decimals but not fraction.")
            self.enable_generate_button()
            return

        # Update all the parameters to the barcode generator
        self.barcode_generator.barcode_type = barcode_type
        self.barcode_generator.frame_type = frame_type
        self.barcode_generator.color_metric = color_metric
        self.barcode_generator.sampled_frame_rate = sampled_frame_rate
        self.barcode_generator.skip_over = skip_over
        self.barcode_generator.total_frames = total_frames

        # Check if user choose the multi-thread or not
        if self.var_multi_thread.get() == 0:
            multi_thread = None
        elif self.var_multi_thread.get() == 1:
            # If user choose to use the multi-thread, then get the number of threads that will be used
            try:
                multi_thread = int(self.thread_entry.get())
                if multi_thread < 1:
                    showwarning("Non Positive Thread Number", "Number of threads has been adjusted to 1.\n"
                                                              "Degenerated to single thread generation.")
                    multi_thread = 1
            except:
                showerror("Invalid Thread Number", "Invalid number of threads.\n"
                                                   "Number of threads must be an integer")
                self.enable_generate_button()
                return

        # Check if user choose to save the frames or not
        if self.var_saved_frame.get() == 1:
            save_frames = True
            try:
                save_frames_rate = int(self.save_frame_entry.get())
            except:
                showerror("Invalid Save Frame Rate", "Invalid Save frame rate.\n"
                                                     "Save frame rate must be a positive Integer.")
                self.enable_generate_button()
        else:
            save_frames_rate = -1
            save_frames = False

        # Check if user choose to rescale the frames or not
        if self.var_rescale_frame.get() == 1:
            try:
                rescale_factor = float(self.rescale_factor_entry.get())
            except:
                showerror("Invalid Rescale Factor", "Invalid frame rescale factor.\n"
                                                    "Must be a positive number.\n"
                                                    "It can be a decimal number but not fractions")
                self.enable_generate_button()
                return
        else:
            rescale_factor = -1

        # Check if user choose to define the letter box region manually
        if self.letterbox_option.get() == "Manual":
            try:
                # Update the letter box parameters, if user choose Manual
                high_ver = int(self.high_ver_entry.get())
                low_ver = int(self.low_ver_entry.get())
                left_hor = int(self.left_hor_entry.get())
                right_hor = int(self.right_hor_entry.get())
                # Start the generation
                self.barcode_generator.generate_barcode(video_filename, user_defined_letterbox=True,
                                                        low_ver=low_ver, high_ver=high_ver,
                                                        left_hor=left_hor, right_hor=right_hor,
                                                        num_thread=multi_thread, save_frames=save_frames,
                                                        rescale_frames_factor=rescale_factor,
                                                        save_frames_rate=save_frames_rate)
            except:
                showwarning("Error Occurred in Barcode Generation", "An unknown Error occurred in the barcode "
                                                                    "generation.\nPlease check the letterbox set up"
                                                                    " and the other parameters' specification.")
                self.enable_generate_button()
                return
        elif self.letterbox_option.get() == "Auto":
            # try:
                # If not, start the generation.
                # The letter box will be automatically found during the generation process
            self.barcode_generator.generate_barcode(video_filename, num_thread=multi_thread,
                                                    save_frames=save_frames,
                                                    rescale_frames_factor=rescale_factor,
                                                    save_frames_rate=save_frames_rate)
            # except:
            #     showwarning("Error Occurred in Barcode Generation", "An unknown Error occurred in the barcode "
            #                                                         "generation.\nPlease check the parameters' "
            #                                                         "specification.")
            #     self.enable_generate_button()
            #     return

        # Correct the total frames
        total_frames = self.barcode_generator.get_barcode().total_frames

        # Get the key of the barcode, which will be later stored in the memory stack (dictionary)
        start_pos = video_filename.rfind("/") + 1
        if start_pos < 0:
            start_pos = 0
        end_pos = video_filename.rfind(".")
        videoname = video_filename[start_pos:end_pos] + "_" + barcode_type + "_" + frame_type + "_" + color_metric \
                    + "_" + str(skip_over) + "_" + str(sampled_frame_rate) + "_" + str(total_frames)

        # Get the barcode from the barcode generator
        barcode = self.barcode_generator.get_barcode()

        # Clear the cv2 captured video object
        barcode.video = None

        # Update the user pre-defined meta data to the computed barcode
        barcode.meta_data = copy.deepcopy(self.meta_data_dict)

        # Add the generated barcode to the memory stack (dictionary)
        self.barcode_stack[videoname] = copy.deepcopy(barcode)

        # Enable the generate button for the next barcode generation request
        self.enable_generate_button()

        # Reset the meta data to the initial state
        self.meta_data_dict = {}

        # Show barcode generation success message
        showinfo("Finished Successfully", "{:s} {:s} {:s} Barcode of the input video:\n"
                                          "{:20s}\n"
                                          "has been successfully generated!\n\n"
                                          "Barcode is saved in the memory with name: {:20s}".format(color_metric,
                                                                                                    frame_type,
                                                                                                    barcode_type,
                                                                                                    video_filename,
                                                                                                    videoname))

    def disable_generate_button(self):
        """
        Disable the generate button and Specify Meta Data button once generation starts
        """
        # Update the generate button text to the processing
        self.generate_button["text"] = "  Processing... "
        self.generate_button.config(state="disabled")
        self.specify_data_button.config(state="disabled")

    def enable_generate_button(self):
        """
        Enable the generate button and Specify Meta Data button in case of failed generation
        """
        # Change the button text back to Generate Barcode
        self.generate_button["text"] = "Generate Barcode"
        self.generate_button.config(state="normal")
        self.specify_data_button.config(state="normal")

    def disable_setup(self):
        """
        Disable the letter box setup entry if user choose the Auto radio button
        """
        self.high_ver_entry.config(state="disabled")
        self.low_ver_entry.config(state="disabled")
        self.left_hor_entry.config(state="disabled")
        self.right_hor_entry.config(state="disabled")

    def enable_setup(self):
        """
        Enable the letter box setup entry if user choose the Manual radio button
        """
        self.high_ver_entry.config(state="normal")
        self.low_ver_entry.config(state="normal")
        self.left_hor_entry.config(state="normal")
        self.right_hor_entry.config(state="normal")



""" LoadJsonWindow Class """

import tkinter.filedialog
from tkinter.messagebox import showerror, showinfo
import copy
import os

from kalmus.tkinter_windows.gui_utils import resource_path, update_graph


class LoadJsonWindow():
    """
    loadJsonWindow Class
    GUI window for user to load the barcode from existed json file to replace with the barcode in the main window
    """
    def __init__(self, barcode_generator, barcode_1, barcode_2, axes,
                 canvas, barcode_stack):
        """
        Initialize

        :param barcode_generator: The barcode generator
        :param barcode_1: The barcode 1
        :param barcode_2: The barcode 2
        :param axes: The axes of the plotted figure in the main window
        :param canvas: The canvas of the plotted figure in the main window
        :param barcode_stack: The dictionary that stores all the barcode on memory
        """
        self.barcode_generator = barcode_generator
        self.barcode_1 = barcode_1
        self.barcode_2 = barcode_2

        # Set up the axes and canvas
        self.axes = axes
        self.canvas = canvas

        self.barcode_stack = barcode_stack

        # Initialize the window
        self.window = tkinter.Tk()
        self.window.wm_title("Load JSON Barcode")
        self.window.iconbitmap(resource_path("kalmus_icon.ico"))

        # Label prompt for the file name/path to the json file
        filename_label = tkinter.Label(self.window, text="JSON file path: ")
        filename_label.grid(row=0, column=0, sticky=tkinter.W)

        # Text entry for user to type the file name/path to the json file
        self.filename_entry = tkinter.Entry(self.window, textvariable="", width=40)
        self.filename_entry.grid(row=0, column=1, columnspan=2, sticky=tkinter.W)

        # Label prompt for user to specify the type of the barcode they will load
        barcode_type_label = tkinter.Label(self.window, text="Specify Barcode Type: ")
        barcode_type_label.grid(row=1, column=0, sticky=tkinter.W)

        # The variable that stores the type of barcode
        self.type_variable = tkinter.StringVar(self.window)
        self.type_variable.set("Color")

        # The dropdown menu for user to select the type of the loaded barcode
        dropdown_type = tkinter.OptionMenu(self.window, self.type_variable, "Color", "Brightness")
        dropdown_type.grid(row=1, column=1, sticky=tkinter.W)

        # Button to build/load the barcode using the given json file
        self.button_build_barcode = tkinter.Button(self.window, text="Load", command=self.build_barcode)
        self.button_build_barcode.grid(row=2, column=3, columnspan=1)

        # Button to browse the folder
        self.button_browse_folder = tkinter.Button(self.window, text="Browse", command=self.browse_folder)
        self.button_browse_folder.grid(row=0, column=3)

        # Variable that stores whcih barcode in the main window to replace with
        self.barcode_option = tkinter.StringVar(self.window)
        self.barcode_option.set("Barcode 1")

        # Radio button for selecting which barcode in the main window to replace with
        radio_barcode_1 = tkinter.Radiobutton(self.window, text="Barcode 1", variable=self.barcode_option,
                                            value="Barcode 1", anchor='w')
        radio_barcode_1.grid(row=1, column=2, sticky=tkinter.W)
        radio_barcode_1.select()

        radio_barcode_2 = tkinter.Radiobutton(self.window, text="Barcode 2", variable=self.barcode_option,
                                           value="Barcode 2", anchor='w')
        radio_barcode_2.grid(row=2, column=2, sticky=tkinter.W)

    def browse_folder(self):
        """
        Browse the folder to locate the json file
        """
        # Get the file name from the user selection
        filename = tkinter.filedialog.askopenfilename(initialdir=".", title="Select JSON file",
                                                      filetypes=(("json files", "*.json"), ("txt files", "*.txt"),
                                                                 ("All files", "*.*")))

        # Update the file name to the file name text entry
        self.filename_entry.delete(0, tkinter.END)
        self.filename_entry.insert(0, filename)

    def build_barcode(self):
        """
        Build/load the barcode use the json file
        """
        # Get the file name/path to the json file
        filename = self.filename_entry.get()

        # Check if the filename is given
        if not os.path.exists(filename):
            showerror("JSON File Not Exists", "JSON file not exists.\n"
                                              "Please check the JSON file path.")
            return

        try:
            # Generate the barcode from json file use the barcode generator
            barcode_type = self.type_variable.get()
            self.barcode_generator.generate_barcode_from_json(filename, barcode_type)
        except:
            showerror("Error Occurred in Loading JSON Barcode", "An error occurred in loading the JSON barcode.\n\n"
                                                                "Please make sure the type of Barcode saved\n"
                                                                "in the JSON file is correctly specified.\n"
                                                                "Color or Brightness")
            return

        # Get the name of the json file
        start_pos = filename.rfind("/") + 1
        if start_pos < 0:
            start_pos = 0

        # Use that as the key to the newly built/loaded barcode
        barcode_name = filename[start_pos: filename.rfind(".json")]
        self.barcode_stack[barcode_name] = copy.deepcopy(self.barcode_generator.get_barcode())

        # Get which barcode in the main window to replace with
        which_barcode = self.barcode_option.get()
        if which_barcode == "Barcode 1":
            self.barcode_1.__dict__ = self.barcode_generator.get_barcode().__dict__.copy()
            self.barcode_1.__class__ = self.barcode_generator.get_barcode().__class__
        elif which_barcode == "Barcode 2":
            self.barcode_2.__dict__ = self.barcode_generator.get_barcode().__dict__.copy()
            self.barcode_2.__class__ = self.barcode_generator.get_barcode().__class__

        # Clear the plotted axes in the main window
        self.axes[0][0].cla()
        self.axes[1][0].cla()
        self.axes[0][1].cla()
        self.axes[1][1].cla()

        # Always plotted the barcode with longer width below
        if self.barcode_1.get_barcode().shape[1] > self.barcode_2.get_barcode().shape[1]:
            temp = copy.deepcopy(self.barcode_1)

            self.barcode_1.__dict__ = self.barcode_2.__dict__.copy()
            self.barcode_2.__dict__ = temp.__dict__.copy()

        # Update the graph/plotted figure in the main window
        update_graph(barcode_1=self.barcode_1, barcode_2=self.barcode_2, axes=self.axes)

        # Redraw the main window
        self.canvas.draw()

        # Quit the main window
        self.window.destroy()

        showinfo("Barcode Loaded Successfully", "{:s} Barcode has been successfully loaded into the memory.\n\n"
                                                "Name key in memory: {:20s}".format(barcode_type, barcode_name))



""" MetaInfoWindow Class """

import tkinter
from kalmus.tkinter_windows.meta_info_windows.SpecifyMetaDataWindow import SpecifyMetaDataWindow
from kalmus.tkinter_windows.gui_utils import resource_path

keys = ["Film Title", "Directors", "Country of Origin", "Produced Year", "Genre"]


class MetaInfoWindow():
    """
    MetaInfoWindow Class
    GUI window that displays the meta information of the selected barcode
    """
    def __init__(self, barcode, barcodes_stack):
        """
        Initialize

        :param barcode: The barcode to check with the meta information
        :param barcodes_stack: The dictionary that store all the barcodes on the memory
        """
        self.barcode = barcode
        self.barcodes_stack = barcodes_stack

        # Initialize the window
        self.window = tkinter.Tk()
        self.window.wm_title("Barcode Meta Information")
        self.window.iconbitmap(resource_path("kalmus_icon.ico"))

        # Label prompt for the meta information display
        self.meta_info_label = tkinter.Label(self.window, text="", width=35, bg='white', anchor='w',
                                             justify=tkinter.LEFT)
        self.meta_info_label.grid(row=0, column=0, columnspan=2, sticky=tkinter.W)

        # Update the Meta information text
        self.refresh_text()

        # Button to refresh the displayed meta information of the barcode
        self.refresh_button = tkinter.Button(master=self.window, text="Refresh",
                                             command=self.refresh_text)
        self.refresh_button.grid(row=1, column=0)

        # Button to add/update meta information to the barcode
        self.add_info_button = tkinter.Button(master=self.window, text="Update Meta Info",
                                              command=self.update_meta_info)
        self.add_info_button.grid(row=1, column=1)

    def refresh_text(self):
        """
        Refresh the meta information text displayed for the barcode
        """
        meta_text = "Meta Information of the Barcode:\n"

        # Build the text using the keys and their values stored in the barcodes's meta information dictionary
        if self.barcode.meta_data is not None:
            for key in keys:
                if key in self.barcode.meta_data.keys():
                    format_string = "{: <s}   {:s}\n".format(str(key) + ":", str(self.barcode.meta_data[key]))
                    meta_text += format_string

        film_length, start_time, end_time = self.get_time_str()
        fps = self.barcode.fps
        meta_text += "\n{: <s}   {:s}\n{: <s}  {:s}\n{: <s}  {:s}\n\n{: <s}  {:.1f}"\
            .format("Film Length:", film_length,
                    "Barcode starts at", start_time,
                    "Barcode ends at", end_time,
                    "Frame rate (FPS):", fps)
        self.meta_info_label["text"] = meta_text

    def get_time_str(self):
        """
        Get string for film length in hrs:mins:secs, clip start time in hrs:mins:secs,
        and clip end time in hrs:mins:secs

        :return: Text string for film length, clip start time at film, clip end time at film
        """
        if self.barcode.fps is None:
            self.barcode.fps = 29.8

        film_length_in_secs = round(self.barcode.film_length_in_frames / self.barcode.fps)

        hrs, mins, secs = self.get_hr_min_sec_from_tot_sec(film_length_in_secs)
        film_length_str = "{:02d}:{:02d}:{:02d}".format(hrs, mins, secs)

        barcode_clip_length_secs = round(self.barcode.total_frames
                                         * (self.barcode.sampled_frame_rate / self.barcode.fps))
        clip_start_time_secs = round(self.barcode.skip_over / self.barcode.fps)

        hrs, mins, secs = self.get_hr_min_sec_from_tot_sec(clip_start_time_secs)
        clip_start_time_str = "{:02d}:{:02d}:{:02d}".format(hrs, mins, secs)

        clip_end_time_secs = clip_start_time_secs + barcode_clip_length_secs
        hrs, mins, secs = self.get_hr_min_sec_from_tot_sec(clip_end_time_secs)
        clip_end_time_str = "{:02d}:{:02d}:{:02d}".format(hrs, mins, secs)

        return film_length_str, clip_start_time_str, clip_end_time_str

    def get_hr_min_sec_from_tot_sec(self, tot_seconds):
        """
        Get the equivalent Hours:Minutes:Seconds representation from a given total seconds

        :param tot_seconds: Total seconds
        :return: Converted hours, minutes, and seconds representation Hours:Minutes:Seconds \
                 Hours >= 0, 60 > Minutes >= 0, 60 > Seconds >= 0
        """
        secs = tot_seconds % 60
        mins = int(tot_seconds / 60) % 60
        hrs = int(tot_seconds / 3600) % 60
        return hrs, mins, secs

    def update_meta_info(self):
        """
        Instantiate the SpecifyMetaDataWindow, if the user press the add info button
        """
        SpecifyMetaDataWindow(self.barcode.__dict__["meta_data"], barcode=self.barcode,
                              barcode_stacks=self.barcodes_stack)



import os
from setuptools import setup, find_packages

dir_path = os.path.abspath(os.path.dirname(__file__))
path_to_md = os.path.join(dir_path, "PyPI_README.md")

with open(path_to_md, encoding='utf-8') as f:
    readme = f.read()
f.close()

setup(name='kalmus',
      version='1.3.8a0',
      description='kalmus film color analysis tool',
      keywords='film, color analysis, data visualization',
      long_description=readme,
      long_description_content_type="text/markdown",
      url='https://github.com/KALMUS-Color-Toolkit/KALMUS',
      author='Yida Chen, Eric Faden, Nathan Ryan',
      author_email='yc015@bucknell.edu',
      license='MIT',
      classifiers=[
          "License :: OSI Approved :: MIT License",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8",
          "Development Status :: 5 - Production/Stable",
          "Intended Audience :: Education",
          "Intended Audience :: Science/Research",
          "Operating System :: OS Independent",
          "Topic :: Multimedia :: Video",
          "Topic :: Software Development :: Libraries :: Python Modules",
          "Topic :: Software Development :: User Interfaces"
      ],
      packages=find_packages(),
      python_requires='>=3.7',
      install_requires=['numpy', 'opencv-python', 'scikit-image>=0.16.2', 'matplotlib==3.2.2',
                        'scikit-learn', 'biopython', 'scipy', 'kiwisolver>=1.3.1'],
      entry_points={
          'console_scripts': ['kalmus-gui=kalmus.command_line_gui:main',
                              'kalmus-generator=kalmus.command_line_generator:main'],
      },
      include_package_data=True,
      zip_safe=False)



""" WhichBarcodeInspectWindow Class """

import tkinter
from kalmus.tkinter_windows.plot_barcodes_windows.PlotBarcodeWindow import PlotBarcodeWindow
from kalmus.tkinter_windows.gui_utils import resource_path


class WhichBarcodeInspectWindow():
    """
    WhichBarcodeInspectWindow Class
    GUI window for user to select which barcode to inspect in detail
    """
    def __init__(self, barcode_1, barcode_2, figsize=(6, 4), dpi=100):
        """
        Initialize

        :param barcode_1: The barcode 1
        :param barcode_2: The barcode 2
        :param figsize: The size of the figure plotted in the window
        :param dpi: dpi of the plotted figure
        """
        self.barcode_1 = barcode_1
        self.barcode_2 = barcode_2
        self.figsize = figsize
        self.dpi = dpi

        # Initialize the window
        self.window = tkinter.Tk()
        self.window.wm_title("Which Barcode to Inspect")
        self.window.iconbitmap(resource_path("kalmus_icon.ico"))

        # Barcode option variable
        self.barcode_option = tkinter.StringVar(self.window)
        self.barcode_option.set("Barcode 1")

        # Option radio button
        radio_barcode_1 = tkinter.Radiobutton(self.window, text="Barcode 1", variable=self.barcode_option,
                                              value="Barcode 1")
        radio_barcode_1.grid(row=0, column=0, padx=50)
        radio_barcode_1.select()

        radio_barcode_2 = tkinter.Radiobutton(self.window, text="Barcode 2", variable=self.barcode_option,
                                              value="Barcode 2")
        radio_barcode_2.grid(row=1, column=0, padx=50)

        # Inspect the barcode button
        self.button_inspect = tkinter.Button(master=self.window, text="Inspect", command=self.inspect_barcode)
        self.button_inspect.grid(row=2, column=0, padx=50)

    def inspect_barcode(self):
        """
        Inspect the selected barcode once user press inspect button
        """
        # Get the selection from the user
        which_barcode = self.barcode_option.get()

        # Quit the current window
        self.window.destroy()

        # Instantiate the PlotBarcodeWindow using the user selection
        if which_barcode == "Barcode 1":
            PlotBarcodeWindow(self.barcode_1, figsize=self.figsize, dpi=self.dpi)
        elif which_barcode == "Barcode 2":
            PlotBarcodeWindow(self.barcode_2, figsize=self.figsize, dpi=self.dpi)



""" Visualization Utility """

import numpy as np
import matplotlib.pyplot as plt
from kalmus.utils.artist import get_contrast_matrix_and_labeled_image


def show_color(color, figure_size=(9, 6), title="",
               axis_off=False, save_image=False, file_name="test.png", return_color=False):
    """
    Plot the rgb color.

    :param color: 1D array contains the R, G, and B channel values
    :param figure_size: Size of the plotted figure
    :param title: The title of the plotted color
    :param axis_off: True to set the axis of the plot figure off. False to have the axis on the side of \
    figure.
    :param save_image: True to save the plot figure into the path that user provide. False not to save \
    the plot figure
    :param file_name: The filename of the saved figure. The default path to the saved image is test.png
    :param return_color: Return the numpy array (shape==30 x 30 x 3) of color block if True. Directly plot the color if False
    :return: Return the numpy array of color block if return_color is True.
    """
    # Generate a block image where each pixel of this block image is the plotted RGB color
    clr = np.ones(shape=(30, 30, 3), dtype=np.uint8) * np.uint8(color)

    if return_color:
        return clr

    # Set up the figure
    plt.figure(figsize=figure_size)
    plt.imshow(clr)
    plt.title(title)
    # Whether to set the axis off
    if axis_off:
        plt.axis('off')
    # Whether to save the plot figure
    if save_image:
        plt.savefig(file_name)
    plt.show()


def show_colors_in_sequence(colors, figure_size=(9, 6), title="", axis_off=False,
                            save_image=False, file_name="test.png", horizontal=True,
                            return_color_sequence=False):
    """
    Plot a sequence of RGB colors either horizontally or vertically in line

    :param colors: 2D array of RGB colors. Expected sequence shape==Number of colors x channels (3)
    :param figure_size: the size of the plot figure
    :param title: the title of the plot
    :param axis_off: True to set the axis of the figure off. False to have the axis with the figure
    :param save_image: True to save the plot figure. False do not save figure
    :param file_name: the path of the saved figure
    :param horizontal: True to plot the sequence of colors horizontally, False to plot the sequence of colors \
                       vertically
    :param return_color_sequence: Return the numpy array (shape==30 x (30 * num_colors) x 3) of color sequence if True \
                                  Directly plot the color sequence if False
    """
    # A chain/sequence of color blocks
    clr_chain = []
    # Size of the color blocks in the display chain/sequence
    block_size = 30
    # Append blocks into the color chain
    for i in range(len(colors)):
        color_block = np.ones(shape=(block_size, block_size, 3), dtype=np.uint8) * np.uint8(colors[i])
        clr_chain.append(color_block)

    clr_chain = np.array(np.concatenate(clr_chain))

    # Now the color chain is vertical. To have horizontal color chain, we transpose the sequence to rotate it
    if horizontal:
        clr_chain = clr_chain.transpose([1, 0, 2])

    if return_color_sequence:
        return clr_chain

    plt.figure(figsize=figure_size)
    plt.imshow(clr_chain)
    plt.title(title)
    if axis_off:
        plt.axis("off")
    if save_image:
        plt.savefig(file_name)
    plt.show()


def show_color_matrix(color_2d_array, mode="truncate", figure_size=(9, 4),
                      title="", axis_off=False,
                      save_image=False, file_name="Block.png", return_matrix=False, return_figure=False):
    """
    Show a matrix of colors (color barcodes). Two modes are available "truncate" and "padding"
    "truncate" mode will truncate the last row of the color matrix if it has different len with respect
    to the other rows. "padding" mode will pad white colors to the last row until the last row has the
    same length as the other rows

    :param color_2d_array: the input color matrix. Expected the shape of the color matrix to be row x col x channels \
    channels should 3 (R, G, and B channels of RGB colors)
    :param mode: mode for displaying color matrix. "truncate" or "padding"
    :param figure_size: the size of the figure
    :param title: the title of the plot
    :param axis_off: True to set the axis of the figure off
    :param save_image: True to save the plot figure, False do not save the figure
    :param file_name: the path to the saved figure
    :param return_matrix: True to return the processed color matrix back False not to return anything
    :param return_figure: Return the plotted figure, if true. Directly plot the color matrix if false
    :return: the processed color matrix. Depending on the display mode, the return color matrix will have padding \
    white colors if mode is "padding", or the last row of color matrix will be truncated if mode is "truncate"
    """
    assert len(color_2d_array.shape) == 3, "Only 2d array of three-channel color tuple/array accepted"
    assert color_2d_array.shape[-1] == 3, "The number of channels of color 2d array must be 3 (R, G, B)"
    assert mode == "truncate" or mode == "padding", "Invalid mode for displaying color matrix, two" \
                                                    "modes are available 'truncate' or 'padding'"
    # If the mode is padding
    if mode == "padding":
        # Flag variable False after appending the second column of colors to the color matrix
        first = True
        for color_col in color_2d_array:
            clr_chain = np.ones(shape=(1, 1, 3), dtype=np.uint8) * np.uint8(color_col[0])
            for i in range(len(color_2d_array[0]) - 1):
                try:
                    color = np.ones(shape=(1, 1, 3), dtype=np.uint8) * np.uint8(color_col[i + 1])
                except:
                    color = np.ones(shape=(1, 1, 3), dtype=np.uint8) * 255
                clr_chain = np.append(clr_chain, color, axis=0)
            if first:
                clr_matrix = np.ones(shape=clr_chain.shape, dtype=np.uint8) * 255
                first = False
            clr_matrix = np.append(clr_matrix, clr_chain, axis=1)
    # If the mode is truncate
    elif mode == "truncate":
        # If the last row has different length with the others
        if len(color_2d_array[-1]) != len(color_2d_array[-2]):
            # Truncate the last row
            clr_matrix = np.array(color_2d_array[:-1]).astype("uint8")
        else:
            # Otherwise keep the matrix
            clr_matrix = np.array(color_2d_array).astype("uint8")
        # Transpose the matrix to have a horizontal display
        clr_matrix = clr_matrix.transpose([1, 0, 2])

    if return_matrix:
        return clr_matrix

    fig = plt.figure(figsize=figure_size)
    plt.imshow(clr_matrix)
    plt.title(title)
    if axis_off:
        plt.axis('off')
    if save_image:
        plt.savefig(file_name)
    if return_figure:
        return fig

    plt.show()


def show_colors_in_cube(colors, figure_size=(8, 8), return_figure=False, sampling=-1, return_sampled_colors=False):
    """
    Show a sequence of RGB colors in cubic RGB space (e.g. R axis (x axis), G axis (y axis),
    and B axis (z axis))
    see https://matplotlib.org/3.1.1/gallery/mplot3d/scatter3d.html for more references
    ipympl is required for this function to work in the jupyter notebook

    :param colors: A sequence of colors to display in cubic space
    :param figure_size: the size of the figure
    :param return_figure: Return the plotted figure and axes, if true \
                          Directly plot the cube, if false
    :param sampling: Number of data points plot in the 3D cube. -1 to include all data in the plot
    :param return_sampled_colors: Return the sampled out colors if True. Directly plot the colors in RGB cube if False
    :return: return the figure and axes with plotted figure if return_figure is True
    """
    assert colors.shape[-1] == 3, "The input colors must be a 2D numpy array with RGB colors where " \
                                  "shape of array is number of colors x channels(3)"

    if sampling > 0:
        sample_indices = np.random.choice(np.arange(colors.size // colors.shape[-1]), sampling, replace=False)
        colors = colors[sample_indices]

    # colors is the N*M x 3 version of the image.
    colors = np.concatenate([np.expand_dims(IC, axis=1) for IC in
                             [colors[..., 0].ravel(), colors[..., 1].ravel(), colors[..., 2].ravel()]], axis=1)

    if return_sampled_colors:
        return colors

    fig = plt.figure(figsize=figure_size)
    ax = fig.add_subplot(111, projection='3d')

    if colors.max() > 1:
        ax.scatter(colors[..., 0], colors[..., 1], colors[..., 2], c=colors[..., :].astype("float32") / 255)
    else:
        ax.scatter(colors[..., 0], colors[..., 1], colors[..., 2], c=colors[..., :])

    # Label the axes.
    ax.set_xlabel('Red')
    ax.set_ylabel('Green')
    ax.set_zlabel('Blue')

    if return_figure:
        return fig, ax

    plt.show()


def show_high_contrast_region(image, minimum_segment_size=0.0004, figsize=(6, 4), return_region_image=False):
    """
    Plot the high contrast region (brightness contrast) of the image.
    image is segmented using the gradient based watershed segmentation

    :param image: input color image shape=(row (height), col (width), channel(3))
    :param minimum_segment_size: The minimum size (in relative ratio [0, 1]) for the segments in the segmented image
    :param figsize: the size of the plot figure
    :param return_region_image: Return the image with only high contrast region if True. \
                                Directly plot the image if False
    :return: If return_region_image is True, return the the image with only high contrast region. \
             region_image.shape == image.shape
    """
    contrast_matrix, labeled_image = get_contrast_matrix_and_labeled_image(image,
                                                                           minimum_segment_size=minimum_segment_size)
    high_contrast_region_index = np.sum(contrast_matrix, axis=1).argmax()

    contrast_part = extract_region_with_index(image, high_contrast_region_index + 1, labeled_image)

    if return_region_image:
        return contrast_part

    plt.figure(figsize=figsize)
    plt.imshow(contrast_part)
    plt.show()


def show_low_contrast_region(image, minimum_segment_size=0.0004, figsize=(6, 4), return_region_image=False):
    """
    Plot the low contrast region (brightness contrast) of the image.
    image is segmented using the gradient based watershed segmentation

    :param image: input color image shape=(row (height), col (width), channel(3))
    :param minimum_segment_size: The minimum size (in relative ratio [0, 1]) for the segments in the segmented image
    :param figsize: the size of the plot figure
    :param return_region_image: Return the image with only low contrast region if True. \
                                Directly plot the image if False
    :return: If return_region_image is True, return the the image with only low contrast region. \
             region_image.shape == image.shape
    """
    contrast_matrix, labeled_image = get_contrast_matrix_and_labeled_image(image,
                                                                           minimum_segment_size=minimum_segment_size)
    low_contrast_region_index = np.sum(contrast_matrix, axis=1).argmin()

    contrast_part = extract_region_with_index(image, low_contrast_region_index + 1, labeled_image)

    if return_region_image:
        return contrast_part

    plt.figure(figsize=figsize)
    plt.imshow(contrast_part)
    plt.show()


def extract_region_with_index(image, region_index, labeled_image):
    """
    Helper function that masked the labeled image on the input image, and extract the
    region with corresponding region index in the labeled image from the input image

    :param image: input image shape=(row(height), col(width))
    :param region_index: the extracted region index in the labeled image
    :param labeled_image: the labeled image of the corresponding input image shape=image.shape
    :return: Image with only region with corresponding index on labeled_image.
    """
    extract_part = image.copy()
    # If the extracted region is bright, paint the background with black
    if np.average(image[labeled_image == region_index]) / float(image.max()) > 0.5:
        extract_part[labeled_image != region_index] = np.array([0, 0, 0], dtype=image.dtype)
    # If the extracted region is dark, paint the background with white
    else:
        # Get the max of image
        max = image.max()
        # If the image is 0-1 normalized
        if max <= 1:
            # White == [1, 1, 1]
            max = 1
        # If the image is 0-255 normalzied (uint8)
        else:
            # White == [255, 255, 255]
            max = 255
        extract_part[labeled_image != region_index] = np.array([max, max, max], dtype=image.dtype)

    return extract_part



""" CalibrateBarcodeTimeWindow Class """
import tkinter

from kalmus.tkinter_windows.gui_utils import resource_path


class CalibrateBarcodeTimeWindow():
    """
    CalibrateBarcodeTimeWindow Class
    GUI window for user to recalibrate the time of a barcode
    """

    def __init__(self, barcode, time_label, frame_label, mouse_x, mouse_y):
        """
        Initialize
        :param barcode: The barcode to recalibrate
        :param time_label: the text label for time
        :param frame_label: the text label for the frame index
        :param mouse_x: the x position of the clicked point
        :param mouse_y: the y position of the clicked point
        """
        self.barcode = barcode
        self.time_label = time_label
        self.frame_label = frame_label
        self.x_pos = mouse_x
        self.y_pos = mouse_y

        # Initialize the window
        self.window = tkinter.Tk()
        self.window.wm_title("At this point...")
        self.window.iconbitmap(resource_path("kalmus_icon.ico"))

        # where does the barcode start (collecting frames from the film)
        start_label = tkinter.Label(master=self.window, text="Start at (min:sec):")
        start_label.grid(row=0, column=0)

        self.start_entry = tkinter.Entry(self.window, textvariable="0", width=12)
        self.start_entry.grid(row=0, column=1, columnspan=1, padx=5, pady=1)

        # where does the barcode end (collecting frames from the film)
        end_label = tkinter.Label(master=self.window, text="End at (min:sec):")
        end_label.grid(row=1, column=0)

        self.end_entry = tkinter.Entry(self.window, textvariable="1", width=12)
        self.end_entry.grid(row=1, column=1, columnspan=1, padx=5, pady=1)

        # the fps of the corresponding film
        fps_label = tkinter.Label(master=self.window, text="Frame per second (fps):")
        fps_label.grid(row=2, column=0)

        self.fps_entry = tkinter.Entry(self.window, textvariable="2", width=12)
        self.fps_entry.grid(row=2, column=1, columnspan=1, padx=5, pady=1)

        # update these information
        self.update_button = tkinter.Button(master=self.window, text="Update", command=self.update)
        self.update_button.grid(row=3, column=0, columnspan=2)

    def update(self):
        """
        Update the start time, end time, and film frame rate (fps)
        """
        # Get the fps
        fps = self.fps_entry.get()
        if len(fps) != 0:
            self.barcode.fps = float(fps)

        # Get the start time
        start_str = str(self.start_entry.get())
        if len(start_str) != 0:
            split_pos = start_str.find(":")
            start_frame = int((int(start_str[:split_pos]) * 60 + int(start_str[split_pos + 1:])) * self.barcode.fps)
            # Update the skip over
            self.barcode.skip_over = start_frame
        else:
            start_frame = self.barcode.skip_over

        # Get the end time
        end_str = str(self.end_entry.get())
        if len(end_str) != 0:
            split_pos = end_str.find(":")
            total_frames = (int(end_str[:split_pos]) * 60 + int(end_str[split_pos + 1:])) * self.barcode.fps
            total_frames -= start_frame
            total_frames = int(total_frames)
        else:
            total_frames = self.barcode.get_barcode().shape[0] * self.barcode.get_barcode().shape[1] \
                           * self.barcode.sampled_frame_rate

        # Compute the scale factor
        self.barcode.scale_factor = total_frames / (self.barcode.get_barcode().shape[0] *
                                                    self.barcode.get_barcode().shape[1] *
                                                    self.barcode.sampled_frame_rate)

        # Recompute the frame label
        frame = (self.barcode.skip_over + self.barcode.sampled_frame_rate * \
                ((self.x_pos * self.barcode.get_barcode().shape[0]) + self.y_pos)) * self.barcode.scale_factor
        frame = int(frame)

        self.frame_label['text'] = "Frame: {:d} ".format(frame)

        # Recompute the time label
        time_tot_sec = frame / self.barcode.fps
        time_sec = int(time_tot_sec % 60)
        time_min = int((time_tot_sec / 60) % 60)
        time_hr = int(time_tot_sec / 3600)

        self.time_label['text'] = "Time: {:02d}:{:02d}:{:02d} ".format(time_hr, time_min, time_sec)

        self.window.destroy()



""" Image Comparison Utility """

import Bio.pairwise2 as sequence_align
import numpy as np
from skimage.color import rgb2hsv
from skimage.metrics import mean_squared_error, structural_similarity


def nrmse_similarity(image_1, image_2, norm_mode="Min max"):
    """
    Normalized root mean squared error (NRMSE).

    :param image_1: The image 1 for comparison
    :param image_2: The image 2 for comparison
    :param norm_mode: The mode for the normalization, average mode use the max (||image_1||, ||image_2||) \
                 Min max use the max(image_1 value range, image_2 value range)
    :return: The score that measure the similarity between two images in range [0,1] using NRMSE \
             0 is the least similar, 1 is the most similar (same)
    """
    image_1 = image_1.astype("float64")
    image_2 = image_2.astype("float64")
    if norm_mode == "Average norm":
        image_1_avg_norm = np.sqrt(np.mean(image_1 * image_1))
        image_2_avg_norm = np.sqrt(np.mean(image_2 * image_2))
        denom = max(image_1_avg_norm, image_2_avg_norm)
    elif norm_mode == "Min max":
        image_1_min_max = image_1.max() - image_1.min()
        image_2_min_max = image_2.max() - image_1.min()
        denom = max(image_1_min_max, image_2_min_max)

    score = 1 - np.sqrt(mean_squared_error(image_1, image_2)) / denom

    return score


def ssim_similarity(image_1, image_2, window_size=None):
    """
    Structural similarity index measure (ssim)

    :param image_1: The image 1 for comparison
    :param image_2: The image 2 for comparison
    :param window_size: The size of the local window, integer
    :return: The Structural similarity index score in range [0,1] \
             0 is the least similar, 1 is the most similar (same)
    """
    assert image_1.shape == image_2.shape, "The shape of two images used for computing structural similarity must " \
                                           "be the same."
    assert len(image_1.shape) >= 2, "The image must be a 2D image (single channel greyscale image or multi-channel" \
                                    "color image)"
    if window_size is not None:
        assert window_size % 2 == 1 and window_size < min(image_1.shape[0], image_1.shape[1]), \
            "The size of the local window must be an odd number and smaller than the size of input images"
    image_1 = image_1.astype("float64")
    image_2 = image_2.astype("float64")

    if len(image_1.shape) == 2:
        score = structural_similarity(image_1, image_2, win_size=window_size, multichannel=False)
    elif len(image_1.shape) > 2:
        score = structural_similarity(image_1, image_2, win_size=window_size, multichannel=True)

    # Renormalize [-1, 1] score to [0, 1] range
    score += 1
    score /= 2

    return score


def get_resample_index(num_frames, sample_amount=10):
    """
    Helper function
    Get the resample indexes based on the number of frames in sequences and the amount of samples we want to
    extract. The indexes are equally spaced. (linear interpolation)

    :param num_frames: The total number of frames
    :param sample_amount: How many frames that you want to sample from them
    :return: np.array of indexes that are equally spaced from 0. The size of the array == sample_amount
    """
    assert num_frames >= sample_amount, "The number of data in"

    possible_index = np.arange(0, num_frames, (num_frames - 1) / (sample_amount - 1))

    nearest_int_index = np.round(possible_index)
    nearest_int_index[-1] = num_frames - 1

    return nearest_int_index.astype('int64')


def cross_correlation(signal_template, signal_source):
    """
    Signal matching. Cross correlation of two input signals. Signals need to be in the same shape

    :param signal_template: The template signal
    :param signal_source: The source signal
    :return: The cross correlation between two input signals. High cross correlation means high similarity between \
             two input signals. range in [-1, 1]
    """
    assert signal_template.shape == signal_source.shape, "The shape of two input signals/color barcodes must have the" \
                                                         "same shapes."
    template = signal_template.copy().astype("float64")
    source = signal_source.copy().astype("float64")
    template -= np.mean(signal_template, axis=tuple(np.arange(len(signal_template.shape) - 1)))
    source -= np.mean(signal_source, axis=tuple(np.arange(len(signal_template.shape) - 1)))
    nom = np.sum(template * source)
    denom = np.sqrt(np.sum(template * template)) * np.sqrt(np.sum(source * source))
    cross_corre = nom / denom

    return cross_corre


def local_cross_correlation(signal_template, signal_source, horizontal_interval=40, vertical_interval=40):
    """
    Local cross correlation between two input signals. The input signals need to be 2 dimensional for local windowing

    :param signal_template: The template signal
    :param signal_source: The source signal
    :param horizontal_interval: Number of horizontal intervals (window width == signal width // horizontal intervals)
    :param vertical_interval: Number of vertical intervals (window height == signal height // vertical intervals)
    :return: The local cross correlation between two signals. Higher local cross correlation means higher similarity \
             between two signals. range in [-1, 1]
    """
    assert signal_source.shape == signal_template.shape, "Incompatiable shape between source and template signals"
    assert len(signal_source.shape) >= 2, "local cross correlation requires the input signals to be 2 dimensional"
    interval_row = signal_template.shape[0] // vertical_interval
    interval_col = signal_template.shape[1] // horizontal_interval

    if interval_row == 0:
        interval_row = 1
    if interval_col == 0:
        interval_col = 1

    template = signal_template.copy().astype("float64")
    source = signal_source.copy().astype("float64")
    for start_row in range(0, template.shape[0], interval_row):
        for start_col in range(0, template.shape[1], interval_col):
            template[start_row: start_row + interval_row, start_col: start_col + interval_col, ...] -= \
                np.mean(template[start_row: start_row + interval_row, start_col: start_col + interval_col, ...],
                        axis=(0, 1))
            source[start_row: start_row + interval_row, start_col: start_col + interval_col, ...] -= \
                np.mean(source[start_row: start_row + interval_row, start_col: start_col + interval_col, ...],
                        axis=(0, 1))

    nom = np.sum(template * source)
    denom = np.sqrt(np.sum(template ** 2)) * np.sqrt(np.sum(source ** 2))
    cross_corre = nom / denom

    return cross_corre


def generate_hue_strings_from_color_barcode(color_barcode, num_interval=12):
    """
    Helper function
    Generate the characters strings that represent the hue values of the input RGB color barcode (3 channel in range
    [0, 255]).

    :param color_barcode: Input color barcode, the input barcode must be a 1 dimensional color barcode with \
                          three channels (R, G, B). shape == [number of colors, 3]
    :param num_interval: The number of intervals that will be divided in the Hue ring (0 to 360 degree)
    :return: The string where each character represent the hue interval of the colors in the input RGB barcode
    """
    assert len(color_barcode.shape) == 2 and color_barcode.shape[-1] == 3, "The input color barcode must be a " \
                                                                           "2D array of 3-chanel RGB colors"
    color_barcode = rgb2hsv(color_barcode.reshape(-1, 1, 3)).reshape(-1, 3)
    hue_barcode = color_barcode[..., 0] * 360
    hue_barcode += 15
    hue_barcode[hue_barcode >= 360] -= 360

    interval_size = 360 / num_interval
    hue_barcode /= interval_size
    hue_barcode = hue_barcode.astype("uint16")

    string_barcode = ""

    for i in hue_barcode:
        str_code = i
        if str_code > 9:
            str_code = chr(ord("a") + (str_code - 9))
        else:
            str_code = str(str_code)
        string_barcode += str_code

    return string_barcode


def generate_brightness_string_from_brightness_barcode(brightness_barcode, num_interval=15):
    """
    Helper function
    Generate the string where each character represents the brightness interval of the brightness in the input
    brightness barcode.

    :param brightness_barcode: Input 1 dimensional brightness barcode with 1 channel. \
                               shape == [number of brightness, 1]
    :param num_interval: The number of intervals that will be divided in the brightness range [0, 255]
    :return: The string where each character represents the brightness interval of the brightness in the input
    """
    assert len(brightness_barcode.shape) == 2 and brightness_barcode.shape[-1] == 1, \
        "The input brightness barcode must be a 2D array with last channel to be 1"
    interval_size = 255 / num_interval
    bri_barcode = brightness_barcode[:, 0] // interval_size
    bri_barcode = bri_barcode.astype("uint16")

    string_barcode = ""
    for i in bri_barcode:
        str_code = i
        if str_code > 9:
            str_code = chr(ord("a") + (str_code - 9))
        else:
            str_code = str(str_code)
        string_barcode += str_code

    return string_barcode


def compare_needleman_wunsch(str_barcode_1, str_barcode_2, local_sequence_size=2000,
                             match_score=2, mismatch_penal=-1, gap_penal=-0.5, extending_gap_penal=-0.1,
                             normalized=False):
    """
    Compare two input character arrays/strings (barcode)'s matching score using the Needleman Wunsch method.
    Needleman Wunsch: https://www.sciencedirect.com/science/article/abs/pii/0022283670900574?via%3Dihub

    :param str_barcode_1: The input string representation of barcode 1
    :param str_barcode_2: The input string representation of barcode 2
    :param local_sequence_size: Divide the long barcode into several small barcode with local_sequence_size length
    :param match_score: The score (bonus) for correctly matching character
    :param mismatch_penal: The penalty for mismatch character
    :param gap_penal: The penalty for gaps within matched sequence
    :param extending_gap_penal: The penalty for extending gaps
    :param normalized: Whether to normalize the final matching score into range [0, 1]
    :return: The match score/normalized match score
    """
    assert len(str_barcode_1) == len(str_barcode_2), "The lengths of two barcodes have to be identical"

    scores = 0
    for start_point in range(0, len(str_barcode_1), local_sequence_size):
        scores += sequence_align.align.globalms(str_barcode_1[start_point:start_point + local_sequence_size],
                                                str_barcode_2[start_point:start_point + local_sequence_size],
                                                match_score, mismatch_penal, gap_penal, extending_gap_penal,
                                                score_only=True)

    if normalized:
        denom = len(str_barcode_1) * match_score
    else:
        denom = 1

    return scores / denom


def compare_smith_waterman(str_barcode_1, str_barcode_2, local_sequence_size=2000,
                           match_score=2, mismatch_penal=-1, gap_penal=-0.5, extending_gap_penal=-0.1,
                           normalized=False):
    """
    Compare two input character arrays/strings (barcode)'s matching score using the Smith Waterman method.
    Smith Waterman: https://www.sciencedirect.com/science/article/abs/pii/0022283681900875?via%3Dihub

    :param str_barcode_1: The input string representation of barcode 1
    :param str_barcode_2: The input string representation of barcode 2
    :param local_sequence_size: Divide the long barcode into several small barcode with local_sequence_size length
    :param match_score: The score (bonus) for correctly matching character
    :param mismatch_penal: The penalty for mismatch character
    :param gap_penal: The penalty for gaps within matched sequence
    :param extending_gap_penal: The penalty for extending gaps
    :param normalized: Whether to normalize the final matching score into range [0, 1]
    :return: The match score/normalized match score
    """
    assert len(str_barcode_1) == len(str_barcode_2), "The lengths of two barcodes have to be identical"

    scores = 0
    for start_point in range(0, len(str_barcode_1), local_sequence_size):
        scores += sequence_align.align.localms(str_barcode_1[start_point:start_point + local_sequence_size],
                                               str_barcode_2[start_point:start_point + local_sequence_size],
                                               match_score, mismatch_penal, gap_penal, extending_gap_penal,
                                               score_only=True)

    if normalized:
        denom = len(str_barcode_1) * match_score
    else:
        denom = 1

    return scores / denom
